# -*- coding: utf-8 -*-
r"""
Module contains classes which belong to testing the precision of two spins from different layers.
"""
from tests.test_multilayers.itest import ITest
import python3.testsignatures as sign
from python3.write_inputs.cwrite_inputs import CWriteInput
from python3.shell_commands import call_spin, change_directory, writelog, cluster, call_sbatch
import time
from tests.test_multilayers.test_precission_withoutdamping.canalyzeprecisionwithoutdamping import \
    CAnalyzePrecisionWithoutDamping
from pathlib import Path


class CTestPrecisionWithoutDamping(ITest):
    r"""
    Test the precision of two spins from different layers.
    """
    @property
    def testsignature(self) -> int:
        r"""
        Returns:
            the testsignature of a test
        """
        return sign.MULTI_PRECISIONWITHOUTDAMPING

    def __init__(self, logfile: Path, cleanup: bool = True, interlayerinteraction: str = 'ab') -> None:
        r"""
        Initializes the test.

        Args:
            interlayerinteraction(str): Interlayer interaction type. It is possible to test within the same unit cell
            (J_ab, J_ac, J_bc) and between different unit cells (J_AA, J_AB, J_AC, J_BA, J_BB, J_BC, J_CA, J_CB, J_CC).
        """
        super().__init__(logfile=logfile, cleanup=cleanup)
        self._interaction = interlayerinteraction
        # interlayer exchange coupling in meV
        self._J = 1.0
        self.setup_input_files()
        self._analyzer = CAnalyzePrecisionWithoutDamping(logfile=logfile, J=self._J, tempdir=self.tempdir,
                                                         lowerline=self.lowerline, upperline=self.upperline)

    def __call__(self) -> bool:
        r"""
        Call the test:
        - Start the calculation
        - Call the analyzer
        - Do the clean up
        """
        with change_directory(self.tempdir):
            if not cluster():
                call_spin()
            else:
                call_sbatch('/work_beegfs/supas384/jobs/job_cau_agheinze.sh')
        while not (self.tempdir / 'SpinSTM_end.dat').is_file():
            time.sleep(0.2)

        l_result = self._analyzer()
        if l_result:
            writelog(self.logfile, message='Test Passed')
        else:
            writelog(self.logfile, message='Test Failed')
        writelog(self.logfile, '=' * 100)

        if self.cleanup:
            self.cleanup_test()

        return l_result

    def setup_input_files(self) -> None:
        r"""
        Setup the input files
        """
        self.write_inp()
        self.write_dyna()
        self.write_lattice()
        self.write_interlayer()
        self.write_simu()
        self.write_stmi()

    def write_stmi(self) -> None:
        r"""
        Writes the SpinSTMi.dat file
        """
        stmi = CWriteInput(file=self.tempdir / 'SpinSTMi.dat')
        stmi()
        if self._interaction in ['ab', 'AA']:
            stmi.appendline(key='', value=' 0.0 0.0 0.0 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 1.0 1.0 0.0 0.0 3.0')
            self.lowerline = 0
            self.upperline = 1
        elif self._interaction == 'bc':
            stmi.appendline(key='', value=' 0.0 0.0 0.0 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 0.5 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 1.0 1.0 0.0 0.0 3.0')
            self.lowerline = 1
            self.upperline = 2
        elif self._interaction == 'ac':
            stmi.appendline(key='', value=' 0.0 0.0 0.0 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 0.5 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 1.0 1.0 0.0 0.0 3.0')
            self.lowerline = 0
            self.upperline = 2
        elif self._interaction == 'AB':
            stmi.appendline(key='', value=' 0.0 0.0 0.0 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 0.5 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 1.0 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 1.5 1.0 0.0 0.0 3.0')
            self.lowerline = 0
            self.upperline = 3
        elif self._interaction == 'BA':
            stmi.appendline(key='', value=' 0.0 0.0 0.0 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 0.5 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 1.0 1.0 0.0 0.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 1.5 0.0 0.0 1.0 3.0')
            self.lowerline = 1
            self.upperline = 2
        elif self._interaction == 'BB':
            stmi.appendline(key='', value=' 0.0 0.0 0.0 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 0.5 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 1.0 1.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 1.5 1.0 0.0 0.0 3.0')
            self.lowerline = 1
            self.upperline = 3
        elif self._interaction == 'AC':
            stmi.appendline(key='', value=' 0.0 0.0 0.0 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 0.3 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 0.6 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 1.0 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 1.3 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 1.6 1.0 0.0 0.0 3.0')
            self.lowerline = 0
            self.upperline = 5
        elif self._interaction == 'BC':
            stmi.appendline(key='', value=' 0.0 0.0 0.0 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 0.3 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 0.6 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 1.0 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 1.3 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 1.6 1.0 0.0 0.0 3.0')
            self.lowerline = 1
            self.upperline = 5
        elif self._interaction == 'CA':
            stmi.appendline(key='', value=' 0.0 0.0 0.0 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 0.3 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 0.6 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 1.0 1.0 0.0 0.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 1.3 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 1.6 0.0 0.0 1.0 3.0')
            self.lowerline = 2
            self.upperline = 3
        elif self._interaction == 'CB':
            stmi.appendline(key='', value=' 0.0 0.0 0.0 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 0.3 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 0.6 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 1.0 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 1.3 1.0 0.0 0.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 1.6 0.0 0.0 1.0 3.0')
            self.lowerline = 2
            self.upperline = 4
        elif self._interaction == 'CC':
            stmi.appendline(key='', value=' 0.0 0.0 0.0 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 0.3 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 0.6 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 1.0 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 1.3 0.0 0.0 1.0 3.0')
            stmi.appendline(key='', value=' 0.0 0.0 1.6 1.0 0.0 0.0 3.0')
            self.lowerline = 2
            self.upperline = 5

    def write_dyna(self) -> None:
        r"""
        Writes the dyna.in file
        """
        dyna = CWriteInput(file=self.tempdir / 'dyna.in', integration='6', timestep='0.1', duration='10000',
                           damping='0.0')
        dyna()

    def write_lattice(self) -> None:
        r"""
        Writes the lattice.in file
        """
        if self._interaction == 'ab':
            lattice = CWriteInput(file=self.tempdir / 'lattice.in', Nsize='1 1 1', alat='1.0 1.0 1.0')
            lattice()
            lattice.appendmultilinevalues(key='lattice', values=['0.5 -0.86602540378 0.0', '0.5 0.86602540378 0.0',
                                                                 '0.0 0.0 1.0'])
            lattice.appendmultilinevalues(key='motif 2 atoms', values=['0.0 0.0 0.0 3.0', '0.0 0.0 1.0 3.0'])
        elif self._interaction == 'AA':
            lattice = CWriteInput(file=self.tempdir / 'lattice.in', Nsize='1 1 2', alat='1.0 1.0 1.0')
            lattice()
            lattice.appendmultilinevalues(key='lattice', values=['0.5 -0.86602540378 0.0', '0.5 0.86602540378 0.0',
                                                                 '0.0 0.0 1.0'])
            lattice.appendmultilinevalues(key='motif 1 atoms', values=['0.0 0.0 0.0 3.0'])
        elif self._interaction in ['ac','bc']:
            lattice = CWriteInput(file=self.tempdir / 'lattice.in', Nsize='1 1 1', alat='1.0 1.0 1.0')
            lattice()
            lattice.appendmultilinevalues(key='lattice', values=['0.5 -0.86602540378 0.0', '0.5 0.86602540378 0.0',
                                                                 '0.0 0.0 1.0'])
            lattice.appendmultilinevalues(key='motif 3 atoms', values=['0.0 0.0 0.0 3.0', '0.0 0.0 0.5 3.0',
                                                                       '0.0 0.0 1.0 3.0'])
        elif self._interaction in ['AB', 'BA', 'BB']:
            lattice = CWriteInput(file=self.tempdir / 'lattice.in', Nsize='1 1 2', alat='1.0 1.0 1.0')
            lattice()
            lattice.appendmultilinevalues(key='lattice', values=['0.5 -0.86602540378 0.0', '0.5 0.86602540378 0.0',
                                                                 '0.0 0.0 1.0'])
            lattice.appendmultilinevalues(key='motif 2 atoms', values=['0.0 0.0 0.0 3.0', '0.0 0.0 0.5 3.0'])
        elif self._interaction in ['AC', 'BC', 'CA', 'CB', 'CC']:
            lattice = CWriteInput(file=self.tempdir / 'lattice.in', Nsize='1 1 2', alat='1.0 1.0 1.0')
            lattice()
            lattice.appendmultilinevalues(key='lattice', values=['0.5 -0.86602540378 0.0', '0.5 0.86602540378 0.0',
                                                                 '0.0 0.0 1.0'])
            lattice.appendmultilinevalues(key='motif 3 atoms', values=['0.0 0.0 0.0 3.0', '0.0 0.0 0.3 3.0',
                                                                       '0.0 0.0 0.6 3.0'])

    def write_inp(self) -> None:
        r"""
        Writes the inp file
        """
        inp = CWriteInput(file=self.tempdir / 'inp', H_ext='0.0 0.0 0.0', J_1='0.0d-3', Periodic_log='.T. .T. .F.',
                          Gra_log='.T. 100')
        inp()

    def write_interlayer(self) -> None:
        r"""
        Writes the interlayer.in file
        """
        interlayer = CWriteInput(file=self.tempdir / 'interlayer.in')
        if self._interaction in ['ab','AA']:
            interlayer.appendline(key='J_' + self._interaction + '1', value=str(self._J) + 'd-3')
            interlayer()
        elif self._interaction == 'bc':
            interlayer.appendline(key='J_ab1', value='0.0d-3')
            interlayer.appendline(key='J_ac1', value='0.0d-3')
            interlayer.appendline(key='J_' + self._interaction + '1', value=str(self._J) + 'd-3')
            interlayer()
        elif self._interaction == 'ac':
            interlayer.appendline(key='J_ab1', value='0.0d-3')
            interlayer.appendline(key='J_bc1', value='0.0d-3')
            interlayer.appendline(key='J_' + self._interaction + '1', value=str(self._J) + 'd-3')
            interlayer()
        elif self._interaction == 'AB':
            interlayer.appendline(key='J_' + self._interaction + '1', value=str(self._J) + 'd-3')
            interlayer.appendline(key='J_ab1', value='0.0d-3')
            interlayer.appendline(key='J_AA1', value='0.0d-3')
            interlayer.appendline(key='J_BB1', value='0.0d-3')
            interlayer.appendline(key='J_BA1', value='0.0d-3')
            interlayer()
        elif self._interaction == 'BA':
            interlayer.appendline(key='J_' + self._interaction + '1', value=str(self._J) + 'd-3')
            interlayer.appendline(key='J_ab1', value='0.0d-3')
            interlayer.appendline(key='J_AA1', value='0.0d-3')
            interlayer.appendline(key='J_BB1', value='0.0d-3')
            interlayer.appendline(key='J_AB1', value='0.0d-3')
            interlayer()
        elif self._interaction == 'BB':
            interlayer.appendline(key='J_' + self._interaction + '1', value=str(self._J) + 'd-3')
            interlayer.appendline(key='J_ab1', value='0.0d-3')
            interlayer.appendline(key='J_AA1', value='0.0d-3')
            interlayer.appendline(key='J_BA1', value='0.0d-3')
            interlayer.appendline(key='J_AB1', value='0.0d-3')
            interlayer()
        elif self._interaction == 'AC':
            interlayer.appendline(key='J_' + self._interaction + '1', value=str(self._J) + 'd-3')
            interlayer.appendline(key='J_ab1', value='0.0d-3')
            interlayer.appendline(key='J_bc1', value='0.0d-3')
            interlayer.appendline(key='J_ac1', value='0.0d-3')
            interlayer.appendline(key='J_AA1', value='0.0d-3')
            interlayer.appendline(key='J_AB1', value='0.0d-3')
            interlayer.appendline(key='J_BA1', value='0.0d-3')
            interlayer.appendline(key='J_BB1', value='0.0d-3')
            interlayer.appendline(key='J_BC1', value='0.0d-3')
            interlayer.appendline(key='J_CA1', value='0.0d-3')
            interlayer.appendline(key='J_CB1', value='0.0d-3')
            interlayer.appendline(key='J_CC1', value='0.0d-3')
            interlayer()
        elif self._interaction == 'BC':
            interlayer.appendline(key='J_' + self._interaction + '1', value=str(self._J) + 'd-3')
            interlayer.appendline(key='J_ab1', value='0.0d-3')
            interlayer.appendline(key='J_bc1', value='0.0d-3')
            interlayer.appendline(key='J_ac1', value='0.0d-3')
            interlayer.appendline(key='J_AA1', value='0.0d-3')
            interlayer.appendline(key='J_AC1', value='0.0d-3')
            interlayer.appendline(key='J_AB1', value='0.0d-3')
            interlayer.appendline(key='J_BA1', value='0.0d-3')
            interlayer.appendline(key='J_BB1', value='0.0d-3')
            interlayer.appendline(key='J_CA1', value='0.0d-3')
            interlayer.appendline(key='J_CB1', value='0.0d-3')
            interlayer.appendline(key='J_CC1', value='0.0d-3')
            interlayer()
        elif self._interaction == 'CA':
            interlayer.appendline(key='J_' + self._interaction + '1', value=str(self._J) + 'd-3')
            interlayer.appendline(key='J_ab1', value='0.0d-3')
            interlayer.appendline(key='J_bc1', value='0.0d-3')
            interlayer.appendline(key='J_ac1', value='0.0d-3')
            interlayer.appendline(key='J_AA1', value='0.0d-3')
            interlayer.appendline(key='J_AC1', value='0.0d-3')
            interlayer.appendline(key='J_AB1', value='0.0d-3')
            interlayer.appendline(key='J_BA1', value='0.0d-3')
            interlayer.appendline(key='J_BB1', value='0.0d-3')
            interlayer.appendline(key='J_BC1', value='0.0d-3')
            interlayer.appendline(key='J_CB1', value='0.0d-3')
            interlayer.appendline(key='J_CC1', value='0.0d-3')
            interlayer()
        elif self._interaction == 'CB':
            interlayer.appendline(key='J_' + self._interaction + '1', value=str(self._J) + 'd-3')
            interlayer.appendline(key='J_ab1', value='0.0d-3')
            interlayer.appendline(key='J_bc1', value='0.0d-3')
            interlayer.appendline(key='J_ac1', value='0.0d-3')
            interlayer.appendline(key='J_AA1', value='0.0d-3')
            interlayer.appendline(key='J_AC1', value='0.0d-3')
            interlayer.appendline(key='J_AB1', value='0.0d-3')
            interlayer.appendline(key='J_BA1', value='0.0d-3')
            interlayer.appendline(key='J_BB1', value='0.0d-3')
            interlayer.appendline(key='J_BC1', value='0.0d-3')
            interlayer.appendline(key='J_CA1', value='0.0d-3')
            interlayer.appendline(key='J_CC1', value='0.0d-3')
            interlayer()
        elif self._interaction == 'CC':
            interlayer.appendline(key='J_' + self._interaction + '1', value=str(self._J) + 'd-3')
            interlayer.appendline(key='J_ab1', value='0.0d-3')
            interlayer.appendline(key='J_bc1', value='0.0d-3')
            interlayer.appendline(key='J_ac1', value='0.0d-3')
            interlayer.appendline(key='J_AA1', value='0.0d-3')
            interlayer.appendline(key='J_AC1', value='0.0d-3')
            interlayer.appendline(key='J_AB1', value='0.0d-3')
            interlayer.appendline(key='J_BA1', value='0.0d-3')
            interlayer.appendline(key='J_BB1', value='0.0d-3')
            interlayer.appendline(key='J_BC1', value='0.0d-3')
            interlayer.appendline(key='J_CA1', value='0.0d-3')
            interlayer.appendline(key='J_CB1', value='0.0d-3')
            interlayer()

    def write_simu(self) -> None:
        r"""
        Writes the simu.in file
        """
        simu = CWriteInput(file=self.tempdir / 'simu.in', i_sd='.T.')
        simu()

    def __repr__(self) -> str:
        r"""
        Returns:
            representation of test
        """
        return f'Testing precision of two spins coupled with interlayer exchange of type: {self._interaction} ...'



import numpy as np
from python3.visualizations import write_STM
from python3.shell_commands import *
import os
import time
from python3.magnetisations import SpinLattice, MultiLayer


def start_spinD(sp, calc_directory, input_directory, dt, duration):
    if not os.path.isdir(calc_directory):
        make_directory(calc_directory)
    copy_element(input_directory / "inp", calc_directory)
    copy_element(input_directory / "lattice.in", calc_directory)
    copy_element(input_directory / "efield.in", calc_directory)
    copy_element(input_directory / "dyna.in", calc_directory)
    copy_element(input_directory / "job.sh", calc_directory)
    copy_element(input_directory / "simu.in", calc_directory)
    adjust_parameter("timestep", dt, calc_directory / "dyna.in")
    adjust_parameter("duration", duration, calc_directory / "dyna.in")
    if type(sp) is MultiLayer :
        z_size = sp.number_layers
    else :
        z_size = 1
    adjust_parameter("Nsize", " " +str(sp.size) + " " + str(sp.size) + " " + str(z_size), calc_directory / "lattice.in")
    # set simulation
    adjust_parameter("i_metropolis", '.F.', calc_directory / "simu.in")
    adjust_parameter("i_paratemp", '.F.', calc_directory / "simu.in")
    adjust_parameter("i_sd", '.T.', calc_directory / "simu.in")
    adjust_parameter("i_gneb", '.F.', calc_directory / "simu.in")
    adjust_parameter("i_htst", '.F.', calc_directory / "simu.in")
    adjust_parameter("i_vt", '.F.', calc_directory / "simu.in")
    adjust_parameter("i_min", '.F.', calc_directory / "simu.in")
    with change_directory(calc_directory):
        sp.write_STM()
        print('Spin Dynamic at: ' + str(calc_directory))
        if cluster() :
            adjust_parameter('#SBATCH --job-name=', '#SBATCH --job-name=spinD', 'job.sh')
            call_sbatch("job.sh")
        else :
            call_spin()



def start_minimisation(sp, calc_directory, input_directory):
    if not os.path.isdir(calc_directory) :
        make_directory(calc_directory)
    copy_element(input_directory / "inp", calc_directory)
    copy_element(input_directory / "lattice.in", calc_directory)
    copy_element(input_directory / "efield.in", calc_directory)
    #copy_element(input_directory + "/dyna.in", calc_directory)
    copy_element(input_directory / "job.sh", calc_directory)
    copy_element(input_directory / "simu.in", calc_directory)
    copy_element(input_directory / "minimization.in", calc_directory)
    if type(sp) is MultiLayer :
        z_size = sp.number_layers
    else :
        z_size = 1
    adjust_parameter("Nsize", " " +str(sp.size) + " " + str(sp.size) + " " + str(z_size), calc_directory / "lattice.in")
    # set simulation
    adjust_parameter("i_metropolis", '.F.', calc_directory / "simu.in")
    adjust_parameter("i_paratemp", '.F.', calc_directory / "simu.in")
    adjust_parameter("i_sd", '.F.', calc_directory / "simu.in")
    adjust_parameter("i_gneb", '.F.', calc_directory / "simu.in")
    adjust_parameter("i_htst", '.F.', calc_directory / "simu.in")
    adjust_parameter("i_vt", '.F.', calc_directory / "simu.in")
    adjust_parameter("i_min", '.T.', calc_directory / "simu.in")
    with change_directory(calc_directory):
        sp.write_STM()
        print('Minimisation at: ' + calc_directory)
        if cluster() :
            adjust_parameter('#SBATCH --job-name=', '#SBATCH --job-name=spinD', 'job.sh')
            call_sbatch("job.sh")
        else :
            call_spin()


#===============================================================================
# INPUT: spinlattice object @sl
#        directory for calculation, string: @calcdirectory
#        directory for the input files, string: @input_directory
#OUTPUT: 1d array with energy contributions from all interactions
#===============================================================================
def numeric_energy(sl, calc_directory, input_directory):
    # SUBFUNCTION: read out energy_density_end.dat
    def get_calenergy_data(calc_directory):
        EXC = []
        DMI = []
        ANI = []
        ZEE = []
        DIP = []
        BIQ = []
        FSP = []
        TOT = []
        for n in range(images) :
            exc = dmi = ani = zee = dip = biq = fsp = tot = 0.0
            enerden = calc_directory / 'energy_density_end.dat'
            with open(enerden) as file:
                for line in file:
                    data = np.asarray(line.split())
                    if data[0][0] == '#':
                        continue
                    exc += float(data[2])
                    dmi += float(data[3])
                    ani += float(data[4])
                    zee += float(data[5])
                    dip += float(data[6])
                    biq += float(data[7])
                    fsp += float(data[8])
                    tot += float(data[9])
            EXC.append(exc)
            DMI.append(dmi)
            ANI.append(ani)
            ZEE.append(zee)
            DIP.append(dip)
            BIQ.append(biq)
            FSP.append(fsp)
            TOT.append(tot)
        return EXC, DMI, ANI, ZEE, DIP, BIQ, FSP, TOT
    # start spindynamics for just 1 step with zero time evolution
    dt = 0
    duration=1
    adjust_parameter("CalEnergy", '.T.', input_directory / "inp")
    start_spinD(sl, calc_directory, input_directory, dt, duration)
    # extract energies
    while not os.path.isfile(calc_directory / 'tot_energy_end.dat'):
        time.sleep(0.2)
    data = np.loadtxt(calc_directory / 'tot_energy_end.dat', unpack=True)
    Exc_intra = data[2]
    Exc_inter = data[3]
    Dmi = data[4]
    Ani = data[5]
    Zee = data[6]
    Dip = data[7]
    Biq = data[8]
    Sp4 = data[9]
    Sp3 = data[10]
    Abs = data[11]
    return np.asarray([Abs, Exc_intra, Zee, Ani, Dmi, Sp3, Sp4, Biq, Dip])



def update_lattice(calc_directory, multilayer=False, number_layers=1):
    SpinSTM_end = calc_directory / 'SpinSTM_end.dat'
    while not os.path.isfile(SpinSTM_end):
        time.sleep(1)
    if multilayer :
        sp = MultiLayer(path=SpinSTM_end, number_layers=number_layers)
    else:
        sp = SpinLattice(path=SpinSTM_end)
    return sp



def start_GNEB(sp_init, sp_final, calc_directory, input_directory, images, iterations, use_path=None, HTST=True):
    if not os.path.isfile(calc_directory) :
        make_directory(calc_directory)
    copy_element(input_directory / "inp", calc_directory)
    copy_element(input_directory / "lattice.in", calc_directory)
    copy_element(input_directory / "efield.in", calc_directory)
    copy_element(input_directory / "job.sh", calc_directory)
    copy_element(input_directory / "GNEB.in", calc_directory)
    copy_element(input_directory / "simu.in", calc_directory)
    time.sleep(1)
    adjust_parameter("i_metropolis", '.F.', calc_directory / "simu.in")
    adjust_parameter("i_paratemp", '.F.', calc_directory / "simu.in")
    adjust_parameter("i_sd", '.F.', calc_directory / "simu.in")
    adjust_parameter("i_gneb", '.T.', calc_directory / "simu.in")
    adjust_parameter("i_vt", '.F.', calc_directory / "simu.in")
    adjust_parameter("i_min", '.F.', calc_directory / "simu.in")
    if HTST :
        copy_element(input_directory + "/htst.in", calc_directory)
        adjust_parameter("i_htst", '.T.', calc_directory / "simu.in")
    adjust_parameter("nim", images, calc_directory / "GNEB.in")
    adjust_parameter("mep_itrmax", iterations, calc_directory / "GNEB.in")
    if sp_init.size == sp_final.size :
        if type(sp_init) is MultiLayer :
            if sp_init.number_layers == sp_final.number_layers :
                z_size = sp_init.number_layers
            else :
                print('ERROR: Initial and Final configurations have different number of layers')
                quit()
        else :
            z_size = 1
        adjust_parameter("Nsize", " " +str(sp_init.size) + " " + str(sp_init.size) + " " + str(z_size), calc_directory / "lattice.in")
    else :
        print('ERROR: Initial and Final configurations have different dimensions')
        quit()
    if use_path :
        copy_all_elements(use_path, '.dat', calc_directory)
        adjust_parameter("amp_rnd_path", 0.0, calc_directory / "GNEB.in")
    with change_directory(calc_directory):
        sp_init.write_STM(name='momfile_i')
        sp_final.write_STM(name='momfile_f')
        print('GNEB at: ', calc_directory)
        if cluster() :
            adjust_parameter('#SBATCH --job-name=', '#SBATCH --job-name=GNEB', 'job.sh')
            call_sbatch("job.sh")
        else :
            call_spin()




def start_MF_interface(SL, modes, calc_directory, input_directory, output_frequency=200, type=None):
    # PRIVATE METHOD FOR FINDING THE HIGHEST MF-IMAGE IN A GIVEN DIRECTORY
    def find_highest_image(directory, stepsize):
        number_tmp = 1
        while True :
            number_tmp += stepsize
            filename_tmp = directory / 'spin_'+str(number_tmp).zfill(4)+'.dat'
            if os.path.isfile(filename_tmp):
                number = number_tmp
            else :
                break
        return number
    # PRIVATE METHOD FOR STARTING A SINGLE MF CALCULATION
    def start_MF(SL, mode, steps, calc_directory, input_directory, vector_init=True, jobname='vt'):
        if not os.path.isfile(calc_directory) :
            make_directory(calc_directory)
        copy_element(input_directory / "inp", calc_directory)
        copy_element(input_directory / "lattice.in", calc_directory)
        copy_element(input_directory / "job.sh", calc_directory)
        copy_element(input_directory / "vt.in", calc_directory)
        copy_element(input_directory / "simu.in", calc_directory)
        # adjust input
        adjust_parameter("i_metropolis", '.F.', calc_directory / "simu.in")
        adjust_parameter("i_paratemp", '.F.', calc_directory / "simu.in")
        adjust_parameter("i_sd", '.F.', calc_directory / "simu.in")
        adjust_parameter("i_gneb", '.F.', calc_directory / "simu.in")
        adjust_parameter("i_htst", '.F.', calc_directory / "simu.in")
        adjust_parameter("i_vt", '.T.', calc_directory / "simu.in")
        adjust_parameter("i_min", '.F.', calc_directory / "simu.in")
        adjust_parameter("vec_init_index", mode, calc_directory / "vt.in")
        adjust_parameter("translation_steps", steps, calc_directory / "vt.in")
        if vector_init :
            adjust_parameter("i_vector_init", " .T. \'vector_init.dat\'", calc_directory / "vt.in")
        else :
            adjust_parameter("i_vector_init", " .F. \'vector_init.dat\'", calc_directory / "vt.in")
        # DO ACTUAL CALL TO MF ROUTINE
        with change_directory(calc_directory):
            SL.write_STM()
            print('VecTrans at: ', calc_directory)
            if cluster() :
                set_jobname(jobname, 'job.sh')
                call_sbatch("job.sh")
            else :
                call_spin()

    # EXCECUTION PART OF FUNCTION
    # ADJUST OUTPUT FREQUENCY
    adjust_parameter("Gra_log", ' .T. '+str(output_frequency) , input_directory / "inp")
    # ERROR CHECK
    if type == None:
        print('ERROR: set @type= \'fresh\' or \'further\' ' )
        exit()
    # STARTS A NEW VECTORTRANSLATION SERIES
    elif type == 'fresh':
        for mode in modes :
            tmp_calc_directory = calc_directory / ('MF_mode_'+str(mode))
            make_directory(tmp_calc_directory)
            # MAKE A FIRST RUN OF ONLY 1 STEP TO GET THE VECOR FOR THE MODE
            fs_directory = tmp_calc_directory / 'first_shoot'
            make_directory(fs_directory)
            start_MF(SL, mode, 1, fs_directory, input_directory, vector_init=False, jobname='vt')
            while not os.path.isfile(fs_directory / 'vt_0001.dat'):
                time.sleep(0.2)
            SL_vec = SpinLattice(path=fs_directory / 'vt_0001.dat')

            # START TRANSLATION IN POSITIVE DIRECTION (fwd)
            fwd_directory = tmp_calc_directory / 'fwd_iteration_1'
            make_directory(fwd_directory)
            write_STM(SL_vec, name=fwd_directory / 'vector_init.dat')
            start_MF(SL, mode, 10000, fwd_directory, input_directory, vector_init=True, jobname='MF'+str(mode)+'_fwd')

            # START TRANSLATION IN NEGATIVE DIRECTION (bck)
            SL_vec.mx, SL_vec.my, SL_vec.mz = -SL_vec.mx, -SL_vec.my, -SL_vec.mz
            bck_directory = tmp_calc_directory / 'bck_iteration_1'
            make_directory(bck_directory)
            write_STM(SL_vec, name=bck_directory / 'vector_init.dat')
            start_MF(SL, mode, 10000, bck_directory, input_directory, vector_init=True, jobname='MF'+str(mode)+'_bck')

    # CREATES A NEW ITERATION OF PREVIOUS MF-SERIES. INPUT @SL WILL BE IGNORED
    elif type == 'further':
        for mode in modes :
            tmp_calc_directory = calc_directory / 'MF_mode_'+str(mode)
            if not os.path.isdir(tmp_calc_directory):
                print('ERROR: Directory, in which the previous MF calculation is supposed to be, does not exist: ', str(tmp_calc_directory))
            # CHECK FOR HIGHEST ITERATION
            n = 1
            check_directory = tmp_calc_directory / 'fwd_iteration_'+str(n)
            while os.path.isdir(check_directory):
                n +=1
                check_directory = tmp_calc_directory / 'fwd_iteration_'+str(n)
            # RESTART fwd: GET SpinLattice AND VECTOR FROM LATEST CALCULATION
            old_fwd_directory = tmp_calc_directory / 'fwd_iteration_'+str(n-1)
            print('highest iteration folder found: ', old_fwd_directory)
            old_image_number = find_highest_image(old_fwd_directory, output_frequency)
            old_image_path =  old_fwd_directory / 'spin_'+str(old_image_number).zfill(4)+'.dat'
            old_image_vect =  old_fwd_directory / 'vt_'+str(old_image_number).zfill(4)+'.dat'
            SL_ini = SpinLattice(path=old_image_path)
            SL_vec = SpinLattice(path=old_image_vect)
            # START CALCULATION IN A NEW DIRECTORY
            fwd_directory = tmp_calc_directory / 'fwd_iteration_'+str(n)
            make_directory(fwd_directory)
            write_STM(SL_vec, name=fwd_directory / 'vector_init.dat')
            start_MF(SL_ini, mode, 10000, fwd_directory, old_fwd_directory, jobname='MF'+str(mode)+'_fwd')
            # RESTART bck: GET SpinLattice AND VECTOR FROM LATEST CALCULATION
            old_bck_directory = tmp_calc_directory / 'bck_iteration_'+str(n-1)
            old_image_number = find_highest_image(old_bck_directory, output_frequency)
            old_image_path =  old_bck_directory / 'spin_'+str(old_image_number).zfill(4)+'.dat'
            old_image_vect =  old_bck_directory / 'vt_'+str(old_image_number).zfill(4)+'.dat'
            SL_ini = SpinLattice(path=old_image_path)
            SL_vec = SpinLattice(path=old_image_vect)
            # START CALCULATION IN A NEW DIRECTORY
            bck_directory = tmp_calc_directory / 'bck_iteration_'+str(n)
            make_directory(bck_directory)
            write_STM(SL_vec, name=bck_directory / 'vector_init.dat')
            start_MF(SL_ini, mode, 10000, bck_directory, old_bck_directory, jobname='MF'+str(mode)+'_fwd')



import numpy as np
import sys
from python3.propertyfunctions import *



class MultiLayer():

    def __init__(self, size=None, key=None, number_layers=None, path=None):
        # CHECK THE INPUT GIVEN
        #if size == None :
        #    print('ERROR: enter kwarg @size=symmetric_xy_system_size_of_single_layer')
        #    exit()
        if number_layers == None:
            print('ERROR: enter kwarg @number_layers=total_number_of_layers')
            exit()
        # INITIALISE LAYERS
        self.Layer = [None for _ in range(number_layers)]
        self.number_layers = number_layers
        # HANDLE READING IN A STM FILE
        if path != None:
            # DO THE ACTUAL READING
            data = np.loadtxt(path, usecols = (0,1,2,3,4,5,6), unpack=False)
            # CHECK IF @number_layers MATCHES THE DATA GIVEN
            if len(data)%number_layers != 0 :
                print('Data given to @MultiLayer has not '+str(number_layers)+' layers, as stated')
                exit()
            size = int(np.sqrt(len(data)//number_layers))
            print('Read in '+str(number_layers)+' layers with size '+str(size))
            # PRESET SpinLattice OBJECTS
            for m in range(number_layers):
                # SpinLattices ARE ARBITRARILY INITIALISED WITH @key='hex'. SHOULD INTERFERE WITH ANYTHING
                self.Layer[m] = SpinLattice(size=size, key='hex', magmom=1.)
            for n in range(len(data)):
                index_layer = n%number_layers
                self.Layer[index_layer].X[n//2], self.Layer[index_layer].Y[n//2], self.Layer[index_layer].Z[n//2] = data[n][0], data[n][1], data[n][2]
                self.Layer[index_layer].mx[n//2], self.Layer[index_layer].my[n//2], self.Layer[index_layer].mz[n//2] = data[n][3], data[n][4], data[n][5]
                self.Layer[index_layer].magmom[n//2] = data[n][6]
            print('Finished setting up @Multilayer object from file '+path)


    #==================================================
    # ADDS A LAYER TO THE @MultiLayer CLASS
    # INPUT: -@SL: @SpinLattice OBJECT RESEMBLIMG THE LAYER TO ADD
    #        -@index_layer: INDEX IN THE ARRAY OF LAYERS
    #        -@zpos_uc: z-POSITION OF THE UNITCELL CONTAINING THE LAYER
    #==================================================
    def add_layer(self, SL, index_layer=0, zpos_uc=0.):
        self.Layer[index_layer] = SL
        self.size = SL.size
        # SET z-POSITION TO z-POSITION OF THE CORRESPONDING UNITCELL
        self.Layer[index_layer].Z = [zpos_uc for n in range(self.Layer[index_layer].size**2)]


    def write_STM(self, name='SpinSTMi.dat'):
        with open(name, 'w') as file :
            for n in range(len(self.Layer[0].X)):
                for m in range(self.number_layers):
                    file.write('%.8f' %self.Layer[m].X[n] + '\t'
                            + '%.8f' %self.Layer[m].Y[n] +  '\t'
                            + '%.8f' %self.Layer[m].Z[n] +  '\t'
                            + '%.8f' %self.Layer[m].mx[n] + '\t'
                            + '%.8f' %self.Layer[m].my[n] + '\t'
                            + '%.8f' %self.Layer[m].mz[n] + '\t'
                            + '%.8f' %self.Layer[m].magmom[n] + '\n')





#==================================================
# class of a general spin lattice.
# this is the core class that contains all necessary instances of the lattice itself.
#==================================================
class SpinLattice():
    def __init__(self, size=None, key=None, path=None, magmom=None, uplo=1., AFM=False, a1_0=None, a2_0=None):
        if size and key and magmom:
            self.setDefault(size, key, magmom, uplo, AFM, a1_0, a2_0)
        elif isinstance(path, str) :
            try:
                self.setSTM(path)
            except:
                print('No SpinSTM found at location:'+ path)
                sys.exit()
        else:
            print('Input Error: make shure you choose one of following:\n 1) size=.., key=.., magmom=..\n 2) path=..')
            sys.exit()

    #==============================================
    # initial configuration of FM state
    #==============================================
    def setDefault(self, size, key, magmom, uplo, AFM, a1_0, a2_0):
        self.size = size
        lat = Lattices(key, size, a1_0, a2_0)
        self.X, self.Y, self.Z = lat.X, lat.Y, lat.Z
        self.mx = [0. for n in range(len(self.X))]
        self.my = [0. for n in range(len(self.X))]
        # if @AFM is true, build initial lattice with alternating z-component
        if AFM and key=='quad' :
            self.mz = [uplo*(-1)**(n%2 + n//size) for n in range(len(self.X))]
        elif AFM and key=='hex' :
            print('Input Error: AFM not possible on hexagonal lattice')
            sys.exit()
        # if no @AFM, build usual FM lattice
        else:
            self.mz = [uplo for n in range(len(self.X))]
        self.magmom = [magmom for n in range(len(self.X))]
    #==============================================
    # takes the spin configuration from STM file as initial one
    #==============================================
    def setSTM(self, STMpath):
        data = np.loadtxt(STMpath, usecols = (0,1,2,3,4,5,6), unpack=True)
        self.X, self.Y, self.Z = data[0], data[1], data[2]
        self.mx, self.my, self.mz = data[3], data[4], data[5]
        self.size = int(np.sqrt(len(self.X)))
        self.magmom = data[6]


    #============================================
    # write spin configuration
    # default name is 'SpinSTMi.dat'
    #============================================
    def write_STM(self, name='SpinSTMi.dat'):
        with open(name, 'w') as file :
            for n in range(len(self.X)):
                file.write('%.8f' %self.X[n] + '\t'
                        + '%.8f' %self.Y[n] +  '\t'
                        + '%.8f' %self.Z[n] +  '\t'
                        + '%.8f' %self.mx[n] + '\t'
                        + '%.8f' %self.my[n] + '\t'
                        + '%.8f' %self.mz[n] + '\t'
                        + '%.8f' %self.magmom[n] + '\n')

    #==============================================
    # method to add a skyrmion to the box
    #==============================================
    def addSkyrmion(self, ursprung, vorticity, helicity, c, w, AFM=False, uplo=1, elongation=False, angl_elongation=0., ab_elongation=[1.,1.]):
        mag = Magnetisation_Skyrmion(self.X, self.Y, ursprung, vorticity, helicity, c, w, AFM, uplo, elongation, angl_elongation, ab_elongation)
        for m in range(len(mag.mz)):
            if (uplo==1 and mag.mz[m] < (1.0-5*10**(-4))) or (uplo==-1 and mag.mz[m] > -(1.0-5*10**(-4))):
                self.mx[m] = mag.mx[m]
                self.my[m] = mag.my[m]
                self.mz[m] = mag.mz[m]

    #==============================================
    # method to add a skyrmion to the box
    #==============================================
    def add_chimera(self, ursprung, vorticity, helicity, c, w, AFM=False, uplo=1, sym_chimera=True, angl_chimera=0., elongation=False, angl_elongation=0., ab_elongation=[1.,1.]):
        mag = Magnetisation_Chimera(self.X, self.Y, ursprung, vorticity, helicity, c, w, AFM, uplo, sym_chimera, angl_chimera, elongation, angl_elongation, ab_elongation)
        for m in range(len(mag.mz)):
            if (uplo==1 and mag.mz[m] < (1.0-5*10**(-4))) or (uplo==-1 and mag.mz[m] > -(1.0-5*10**(-4))):
                self.mx[m] = mag.mx[m]
                self.my[m] = mag.my[m]
                self.mz[m] = mag.mz[m]

    #==============================================
    # method to add a meron to the box
    #==============================================
    def add_meron(self, ursprung, vorticity, helicity, c, w, AFM=False, uplo=1):
        mag = Magnetisation_Meron(self.X, self.Y, ursprung, vorticity, helicity, c, w, AFM, uplo)
        for m in range(len(mag.mz)):
            if (uplo==1 and mag.mz[m] < 0.51) or (uplo==-1 and mag.mz[m] > -0.51):
                self.mx[m] = mag.mx[m]
                self.my[m] = mag.my[m]
                self.mz[m] = mag.mz[m]

    #==============================================
    # method to add a bimeron to the box
    #==============================================
    def add_bimeron(self, pos0, R, a1, a2):
        mag = Magnetisation_Bimeron(self.X, self.Y, pos0, R, a1, a2)
        for m in range(len(mag.mz)):
            self.mx[m] = mag.mx[m]
            self.my[m] = mag.my[m]
            self.mz[m] = mag.mz[m]

    #==============================================
    # method to add a skyrmion to the box
    #==============================================
    def addSkyrmionlattice(self, k, vorticity, helicity, c, w, AFM=False, uplo=1, onside=True):
        index_dist = self.size//k
        if onside :
            delta_X = 0.0
            delta_Y = 0.0
        else :
            delta_X = self.X[1]
            delta_Y = self.Y[1]/2.
        # Iterate through all lattice sides (+1 for treating periodic boundaries)
        for n in range(self.size+1):
            for m in range(self.size+1):
                # If lattice side matches periodic kriteria, decide, if periodic
                # boundary conditions apply or not and set the proper position for
                # the skyrmion center
                if n%index_dist == 0 and m%index_dist == 0 :
                    if n >= self.size and m >= self.size :
                        skyrmion_index = (n-1)*self.size + m -1
                        pos = [self.X[skyrmion_index] + self.X[self.size] + self.X[1] + delta_X, self.Y[skyrmion_index]+ self.Y[self.size]+ self.Y[1]+ delta_Y]
                    elif n >= self.size :
                        skyrmion_index = (n-1)*self.size +m
                        pos = [self.X[skyrmion_index] + self.X[self.size] + delta_X, self.Y[skyrmion_index]+ self.Y[self.size]+ delta_Y]
                    elif m >= self.size :
                        skyrmion_index = n*self.size + m -1
                        pos = [self.X[skyrmion_index] + self.X[1] + delta_X, self.Y[skyrmion_index]+ self.Y[1]+ delta_Y]
                    # this is the case, when no periodic boundary conditions have to be considered
                    else :
                        skyrmion_index = n*self.size + m
                        pos = [self.X[skyrmion_index] +delta_X, self.Y[skyrmion_index]+ delta_Y]
                    # PLACE SKYRMION
                    self.addSkyrmion(pos, vorticity, helicity, c, w, AFM=AFM, uplo=uplo)

    #==============================================
    # method to create a 3q state for the whole box
    #==============================================
    def add3qstate(self, init=None):
        mag = Magnetisation_3qstate(self.X, self.Y, init)
        self.mx = mag.mx
        self.my = mag.my
        self.mz = mag.mz
    #==============================================
    # method to create a 1q state for the whole box
    #==============================================
    def add1qstate(self, Rq, Iq, Q, phase=0.0):
        mag = Magnetisation_1qstate(self.X, self.Y, self.Z, Rq, Iq, Q, phase)
        self.Q, self.Rq, self.Iq = mag.Q, mag.Rq, mag.Iq
        self.mx = mag.mx
        self.my = mag.my
        self.mz = mag.mz

    #==============================================
    # method to create a 1q state for the whole box
    #==============================================
    def add_superposition(self, sl1, sl2):
        mag = Magnetisation_superposition(self.X, self.Y, self.Z, sl1, sl2)
        self.mx = mag.mx
        self.my = mag.my
        self.mz = mag.mz
    #==============================================
    # method to add a domainwall to the box
    #==============================================
    def addDomainwall(self, r0, direction, width, heli=np.pi):
        mag = Magnetisation_Domainwall(self.X, self.Y, r0, direction, width, heli)
        self.r0 = r0
        self.direction = np.asarray(direction)/np.linalg.norm(direction)
        self.dw_width = width
        for m in range(len(mag.mz)):
            if mag.mz[m] < (1.0-10**(-2)):
                self.mx[m] = mag.mx[m]
                self.my[m] = mag.my[m]
                self.mz[m] = mag.mz[m]

    #==============================================
    # method to add the row wise antiferromagnet to the whole box
    #==============================================
    def addRWAFM(self, init='inplaneX'):
        mag = Magnetisation_RWAFM(self.X, self.Y, init)
        self.mx = mag.mx
        self.my = mag.my
        self.mz = mag.mz

    #==============================================
    # method to return the topological charge of the lattice
    #==============================================
    def getTopocharge(self):
        mag = [self.mx, self.my, self.mz]
        self.topo_charge = topocharge(mag)
    #==============================================
    # method to return the radius of an isolated skyrmion.
    # if the lattice inhabitats more than one skyrmion or
    # something else, an exception will be raised
    #==============================================
    def getSkradius(self):
        self.sk_radius, self.sk_popt, self.sk_pcov = skyrmionradius(self)
    #==============================================
    # method for finding the width of a domainwall
    #==============================================
    def getDWwidth(self):
        self.r0, self.phi, self.dw_width = dwwidth(self)
        self.direction = np.asarray([np.cos(self.phi), np.sin(self.phi)])






#===============================================================
# class for the construction of the lattice in such a way, that
# the result can be used in the Spindynamics code
# quad or hex can be chosen
#===============================================================
class Lattices():
    def __init__(self, key, size, a1_0=None, a2_0=None):
        if key == 'quad':
            self.X, self.Y, self.Z = self.build_quadratic_lattice(size, a1_0, a2_0)
        elif key == 'hex':
            self.X, self.Y, self.Z = self.build_hexagonal_lattice(size, a1_0, a2_0)
    #==============================================
    # method returns lists @X, @Y, @Z so, that (X[n], Y[n], Z[n])
    # is a point on the hexagonal lattice
    # The kwargs a1_0 and a2_0 define the zero point of the lattice in units of
    # the right handed lattice vectors. These are only important for Multilayers
    #==============================================
    def build_hexagonal_lattice(self, size, a1_0=None, a2_0=None):
        c1 = 0.5
        c2 = -round(np.sqrt(3.0/4.0),8)
        if a1_0 and a2_0:
            x0 = 0.0 + c1 * (a1_0 + a2_0)
            y0 = 0.0  + c2 * (a2_0 - a1_0)
        else:
            x0, y0 = 0.0, 0.0
        X, Y, Z = [], [], []
        for n in range(size):
            for m in range(size):
                X.append(x0 + m*c1)
                Y.append(y0 - m*c2)
                Z.append(0)
            x0 += c1
            y0 += c2
        return X,Y,Z
    #==============================================
    # method returns lists @X, @Y, @Z so, that (X[n], Y[n], Z[n])
    # is a point on the quadratic lattice
    # The kwargs a1_0 and a2_0 define the zero point of the lattice in units of
    # the right handed lattice vectors. These are only important for Multilayers
    #==============================================
    def build_quadratic_lattice(self, size, a1_0=None, a2_0=None):
        if a1_0 and a2_0:
            x0 = 0.0 + a1_0
            y0 = 0.0 + a2_0
        else:
            x0, y0 = 0.0, 0.0
        x = np.linspace(0, size-1, size)
        y = np.linspace(0, size-1, size)
        X, Y, Z = [], [], []
        for n in x:
            for m in y:
                 X.append(n + x0)
                 Y.append(m + y0)
                 Z.append(0)
        return X,Y,Z




#==============================================================
# takes parameters and arrays X,Y and returns diskret magnetisation
# for skyrmions with corresponding structure
#===============================================================
class Magnetisation_Skyrmion():
    def __init__(self, X, Y, pos0, m, g, c, w, AFM, uplo, elongation, angl_elongation, ab_elongation):
        self.mx, self.my, self.mz = self.build_skyrmion(X, Y, pos0, m, g, c, w, uplo, AFM, elongation, angl_elongation, ab_elongation)

    #==========================================================
    # functions for the polar and azimutal angle of an skyrmion.
    # theta is the standard bogdanov profile
    #==========================================================
    def theta(self, r,c,w):
        comp1 = np.arcsin(np.tanh((-r -c)*2/w))
        comp2 = np.arcsin(np.tanh((-r +c)*2/w))
        return np.pi + comp1 + comp2

    def phi(self, p, m, g):
        return  m*p + g
    #==========================================================
    # here the magnetisation of the skyrmion is build with
    # respect to the profile parameters, helicity and vorticity
    #==========================================================
    def build_skyrmion(self, X, Y, pos0, m, g, c, w, uplo, AFM, elongation, angl_elongation, ab_elongation):
        assert len(X) == len(Y)
        mx, my, mz = [], [], []
        for n in range(len(X)):
            x, y = X[n]-pos0[0], Y[n]-pos0[1]
            r, p = np.sqrt(x**2 + y**2), np.arctan2(y,x)
            # HANDLE ELONGATION OF SKYRMION IF CHOSEN. USE ELLIPTIC PARAMETERISATION FOR RADIAL DISTANCE @r
            if elongation :
                #r = r/(1+exz*abs(np.cos(p-angl_elongation)))
                r = r*(2.-np.sqrt((ab_elongation[0]*np.cos(p-angl_elongation))**2 + (ab_elongation[1]*np.sin(p-angl_elongation))**2))

            # SET AZIMUTHAL ANGLE FOR RADIAL THETA PROFILE
            th = self.theta(r,c,w)
            ph = self.phi(p, m, g)
            # alternate sign of spin if @AFM is True
            if AFM :
                sign = (-1)**(n%2 + n//int(np.sqrt(len(X))))
            else :
                sign = 1
            # actually setting of vector components resembling spin
            mx.append(sign*np.sin(th)*np.cos(ph))
            my.append(sign*np.sin(th)*np.sin(ph))
            mz.append(sign*np.cos(th)*float(uplo))
        return mx, my, mz


#==============================================================
# takes parameters and arrays X,Y and returns diskret magnetisation
# for skyrmions with corresponding structure
#===============================================================
class Magnetisation_Chimera():
    def __init__(self, X, Y, pos0, m, g, c, w, AFM, uplo, sym_chimera, angl_chimera, elongation, angl_elongation, ab_elongation):
        self.mx, self.my, self.mz = self.build_chimera(X, Y, pos0, m, g, c, w, uplo, AFM, sym_chimera, angl_chimera, elongation, angl_elongation, ab_elongation)

    #==========================================================
    # functions for the polar and azimutal angle of an skyrmion.
    # theta is the standard bogdanov profile
    #==========================================================
    def theta(self, r,c,w):
        comp1 = np.arcsin(np.tanh((-r -c)*2/w))
        comp2 = np.arcsin(np.tanh((-r +c)*2/w))
        return np.pi + comp1 + comp2

    def phi(self, p, m, g):
        return  m*p + g
    #==========================================================
    # here the magnetisation of the skyrmion is build with
    # respect to the profile parameters, helicity and vorticity
    #==========================================================
    def build_chimera(self, X, Y, pos0, m, g, c, w, uplo, AFM, sym_chimera, angl_chimera, elongation, angl_elongation, ab_elongation):
        assert len(X) == len(Y)
        mx, my, mz = [], [], []
        for n in range(len(X)):
            x, y = X[n]-pos0[0], Y[n]-pos0[1]
            r, p = np.sqrt(x**2 + y**2), np.arctan2(y,x)
            # HANDLE ELONGATION OF SKYRMION IF CHOSEN. USE ELLIPTIC PARAMETERISATION FOR RADIAL DISTANCE @r
            if elongation :
                #r = r/(1+exz*abs(np.cos(p-angl_elongation)))
                r = r*(2.-np.sqrt((ab_elongation[0]*np.cos(p-angl_elongation))**2 + (ab_elongation[1]*np.sin(p-angl_elongation))**2))

            # SET AZIMUTHAL ANGLE FOR RADIAL THETA PROFILE
            th = self.theta(r,c,w)
            # chimera case: change vorticity if @p switches half planes
            # the idea is to map @p to @p_tmp, which is 0<p_tmp<pi/2 for every spin, whos vorticity has to be switched
            # EDIT: IF WE WANT TO HAVE SYMMETRIC CHIMERA STATES, WE HAVE TO DEVIDE NOT ONLY INTO HALF PLANES
            #       INSTEAD WE HAVE TO HAVE INTERVALS OF pi/m , WITH VORTICITY @m
            if sym_chimera :
                plane_devider = m
            else :
                plane_devider = 1
            # MAP POLAR ANGLE TO EASIER TO HANDLE INTERVALL
            p_tmp = p + np.pi/2. -angl_chimera
            #p_tmp = p -angl_chimera
            # SINCE p=arctan(x,y) IS IN [-pi, +pi], SUBTRACTING @angl_chimera CAN VIOLATE THESE BOUNDARIES --> HANDLE THIS BY MAPPING BACK
            if p_tmp < -np.pi :
                p_tmp += 2.*np.pi
            # DECIDE; WHICH VORTICITY TO TAKE, DEPENDING ON INTERVALL WE ARE IN RIGHT NOW
            if (p_tmp > 0) and (p_tmp < np.pi/plane_devider) :
                ph = self.phi(p, -m, g + np.pi + 2*angl_chimera)
            else :
                ph = self.phi(p, m, g)

            # alternate sign of spin if @AFM is True
            if AFM :
                sign = (-1)**(n%2 + n//int(np.sqrt(len(X))))
            else :
                sign = 1
            # actually setting of vector components resembling spin
            mx.append(sign*np.sin(th)*np.cos(ph))
            my.append(sign*np.sin(th)*np.sin(ph))
            mz.append(sign*np.cos(th)*float(uplo))
        return mx, my, mz


#==============================================================
# takes parameters and arrays X,Y and returns diskret magnetisation
# for skyrmions with corresponding structure
#===============================================================
class Magnetisation_Meron():
    def __init__(self, X, Y, pos0, m, g, c, w, AFM, uplo):
        self.mx, self.my, self.mz = self.build_meron(X, Y, pos0, m, g, c, w, uplo)
    #==========================================================
    # functions for the polar and azimutal angle of an skyrmion.
    # theta is the standard bogdanov profile
    #==========================================================
    def theta(self, r,c,w):
        comp1 = np.arcsin(np.tanh((-r -c)*2/w))
        comp2 = np.arcsin(np.tanh((-r +c)*2/w))
        return np.pi + comp1/2. + comp2/2.

    def phi(self, p, m, g):
        return  m*p + g
    #==========================================================
    # here the magnetisation of the skyrmion is build with
    # respect to the profile parameters, helicity and vorticity
    #==========================================================
    def build_meron(self, X, Y, pos0, m, g, c, w, uplo):
        assert len(X) == len(Y)
        mx, my, mz = [], [], []
        for n in range(len(X)):
            x, y = X[n]-pos0[0], Y[n]-pos0[1]
            r, p = np.sqrt(x**2 + y**2), np.arctan2(y,x)
            th = self.theta(r,c,w)
            ph = self.phi(p,m,g)
            mx.append(np.sin(th)*np.cos(ph))
            my.append(np.sin(th)*np.sin(ph))
            mz.append(np.cos(th)*float(uplo))
        return mx, my, mz

#============================================================
# takes the position arrays X, Y, and the Bimeron parameters a1, a2, R
# R --> distance between the Meron centers
# a1, a2 --> paramteters with determine if the Bimeron is roud or elliptical
# a1 = a2 --> round Bimeron
# pos0 --> zero point as array
#============================================================
class Magnetisation_Bimeron():
    def __init__(self, X, Y, pos0, R, a1, a2):
        self.mx, self.my, self.mz = self.build_bimeron(X, Y, pos0, R, a1, a2)

    #===============================================
    #magnetisation direction in spherical coordinates
    #===============================================
    def theta(self, R, c_i, c_j, a1, a2):
        rho = np.sqrt(a1*c_i**2 + a2*c_j**2)
        return np.arccos(R * c_i / (rho**2 + (R**2)/4 ))

    def phi(self, R, c_i, c_j):
        return np.arctan((c_i - R/2)/(c_j)) - np.arctan((c_i + R/2)/(c_j))

    def build_bimeron(self, X, Y, pos0, R, a1, a2):
        r = 1
        assert len(X) == len(Y)
        mx, my, mz = np.zeros(len(X)), np.zeros(len(X)), np.zeros(len(X))
        for n in range(len(X)):
            theta_c = self.theta(R, X[n]-pos0[0], Y[n]-pos0[1], a1, a2)
            phi_c   = self.phi(R, X[n]-pos0[0], Y[n]-pos0[1])
            mx[n] = np.round(r*np.sin(theta_c)*np.cos(phi_c),8)
            my[n] = np.round(r*np.sin(theta_c)*np.sin(phi_c),8)
            mz[n] = np.round(r*np.cos(theta_c),8)
        return mx, my, mz


#==================================================
# class for magnetisation identified with the
# 1q-state also known as simple spin spiral
#==================================================
class Magnetisation_1qstate():

    def __init__(self, X, Y, Z, Rq, Iq, Q, phase):
        # reshape Rq, Iq, Q with respect to lattice vectors
        latvec = np.asarray([ [0.5, -0.86602540378, 0.0],
                            [0.5, 0.86602540378, 0.0],
                            [0.0, 0.0, 1.0] ]).astype('float')
        spat = np.dot(latvec[0], np.cross(latvec[1], latvec[2]))
        Q = ( Q[0]*np.cross(latvec[1], latvec[2])
            + Q[1]*np.cross(latvec[2], latvec[0])
            + Q[2]*np.cross(latvec[0], latvec[1]) )*2.0*np.pi/spat
        self.Q = Q
        Rq = Rq[0]*latvec[0] + Rq[1]*latvec[1] + Rq[2]*latvec[2]
        self.Rq = Rq/np.linalg.norm(Rq)
        Iq = Iq[0]*latvec[0] + Iq[1]*latvec[1] + Iq[2]*latvec[2]
        self.Iq = Iq/np.linalg.norm(Iq)
        # calculate magnetisation
        self.mx, self.my, self.mz = self.build_1qstate(X, Y, Z,
                                                       self.Rq, self.Iq, self.Q, phase)

    def build_1qstate(self, X, Y, Z, Rq, Iq, Q, phase):
        mx, my, mz = [], [], []
        for n in range(len(X)):
            R = np.asarray([X[n], Y[n], Z[n]]).astype('float')
            dot = np.dot(R, Q) + phase
            mx.append(Rq[0]*np.cos(dot) - Iq[0]*np.sin(dot))
            my.append(Rq[1]*np.cos(dot) - Iq[1]*np.sin(dot))
            mz.append(Rq[2]*np.cos(dot) - Iq[2]*np.sin(dot))
        return mx,my,mz


#==================================================
# class for magnetisation identified with the
# 3q-state, definitions from Dissertation Phillip Kurz S.111
#==================================================
class Magnetisation_3qstate():
    def __init__(self, X, Y, init):
        self.mx, self.my, self.mz = self.build_3qstate(X, Y)
        if init == 'ori3':
            self.mx, self.my, self.mz = self.align_ori3(self.mx, self.my, self.mz)
        elif init == 'ori2':
            self.mx, self.my, self.mz = self.align_ori2(self.mx, self.my, self.mz)
        elif init == 'ori1':
            self.mx, self.my, self.mz = self.align_ori1(self.mx, self.my, self.mz)
    #==================================================
    # Input: arrays X, Y, so that (X[i], Y[i]) is a point on
    # the lattice
    # Note: A,B,C can be chosen to generate other multi-q-states
    #==================================================
    def build_3qstate(self, X, Y):
        q1 = 2.0*np.pi*np.asarray([0, 1.0/np.sqrt(3)])
        q2 = 2.0*np.pi*np.asarray([0.5, 1.0/np.sqrt(3)/2])
        q3 = 2.0*np.pi*np.asarray([0.5, -1.0/np.sqrt(3)/2])
        A = 1.0/np.sqrt(3)
        B = 1.0/np.sqrt(3)
        C = 1.0/np.sqrt(3)
        mx, my, mz = [], [], []
        for n in range(len(X)):
            R = np.asarray([X[n], Y[n]])
            mx.append(A*np.cos(np.dot(q1,R)))
            my.append(B*np.cos(np.dot(q2,R)))
            mz.append(C*np.cos(np.dot(q3,R)))
        return mx,my,mz
    #==================================================
    # turns the magnetisation of the initial PK state into
    # highly symmetrical ori3
    #==================================================
    def align_ori3(self, mx, my, mz):
        # 1) rotate spins, so that Spin 0 is aligned with z-axis
        axis1, ang1 = rotate_to_axis([0.0, 0.0, 1.0], [mx[0], my[0], mz[0]])
        mat1 = rotation_matrix(axis1, ang1)
        mx, my, mz = rotate_spins(mx, my, mz, mat1)
        # 2) rotate Spins, so that Spin 1 points along (1/2 +isqrt(3)/2)
        axis2, ang2 = rotate_to_axis([0.5, np.sqrt(3.0)/2.0, 0.0], [mx[1], my[1], 0.0])
        mat2 = rotation_matrix(axis2, ang2)
        mx,my,mz = rotate_spins(mx,my,mz, mat2)
        return mx, my, mz

    def align_ori2(self, mx, my, mz):
        # start with a ori3 configuration
        mx, my, mz = self.align_ori3(mx, my, mz)
        # now rotate by \tau/2 around y-axis
        rotmat = rotation_matrix([0.0, 1.0, 0.0], np.arccos(-1.0/3.0)/2.0)
        mx,my,mz = rotate_spins(mx,my,mz, rotmat)
        return mx, my, mz

    def align_ori1(self, mx, my, mz):
        # start with a ori3 configuration
        mx, my, mz = self.align_ori2(mx, my, mz)
        # now rotate by \tau/2 around y-axis
        rotmat = rotation_matrix([0.0, 1.0, 0.0], -np.pi/2.0)
        mx,my,mz = rotate_spins(mx,my,mz, rotmat)
        return mx, my, mz



#==================================================
# class for magnetisation identified with the
# 1q-state also known as simple spin spiral
#==================================================
class Magnetisation_superposition():

    def __init__(self, X, Y, Z, sl1, sl2):
        self.mx, self.my, self.mz = self.build_superposition(X, Y, Z, sl1, sl2)

    def lattice_fft(self, sp):
        L = sp.size
        xfft = (np.fft.ifft2(np.asarray(sp.mx).reshape(L, L)).transpose()).reshape(sp.size**2)
        yfft = (np.fft.ifft2(np.asarray(sp.my).reshape(L, L)).transpose()).reshape(sp.size**2)
        zfft = (np.fft.ifft2(np.asarray(sp.mz).reshape(L, L)).transpose()).reshape(sp.size**2)
        return xfft, yfft, zfft

    def inverse_lattice_fft(self, fft_x, fft_y, fft_z):
        L = int(np.sqrt(len(fft_x)))
        x = (np.fft.fft2(np.asarray(fft_x).reshape(L, L)).transpose()).reshape(L**2)
        y = (np.fft.fft2(np.asarray(fft_y).reshape(L, L)).transpose()).reshape(L**2)
        z = (np.fft.fft2(np.asarray(fft_z).reshape(L, L)).transpose()).reshape(L**2)
        return x, y, z

    def normalise_inverse_fft(self, x,y,z):
        for n in range(len(x)):
            norm = np.sqrt(x[n]**2 + y[n]**2 + z[n]**2)
            x[n] = x[n]/norm
            y[n] = y[n]/norm
            z[n] = z[n]/norm
        return x,y,z

    def build_superposition(self, X, Y, Z, sl1, sl2):
        mx, my, mz = [], [], []
        # calculate fourier representation
        sl1_x, sl1_y, sl1_z = self.lattice_fft(sl1)
        sl2_x, sl2_y, sl2_z = self.lattice_fft(sl2)
        # build superposition of waves
        sl_x = (sl1_x + sl2_x)*np.sqrt(2.)
        sl_y = (sl1_y + sl2_y)*np.sqrt(2.)
        sl_z = (sl1_z + sl2_z)*np.sqrt(2.)
        # transform back to real space image
        mx, my, mz = self.inverse_lattice_fft(sl_x, sl_y, sl_z)
        mx, my, mz = self.normalise_inverse_fft(mx.real, my.real, mz.real)
        return mx, my, mz



#==================================================
# class for magnetisation identified with the
# 3q-state, definitions from Dissertation Phillip Kurz S.111
#==================================================
class Magnetisation_RWAFM():
    def __init__(self, X, Y, init):
        self.mx, self.my, self.mz = self.build_3qstate(X, Y, init)
    #==================================================
    # Input: arrays X, Y, so that (X[i], Y[i]) is a point on
    # the lattice
    # Note: A,B,C can be chosen to generate other multi-q-states
    #==================================================
    def build_3qstate(self, X, Y, init):
        if init == 'outofplane':
            A, B, C = 0, 0, 1
        elif init == 'inplaneY':
            A, B, C = 0, 1, 0
        elif init == 'inplaneX':
            A, B, C = -1, 0, 0
        q1 = 2.0*np.pi*np.asarray([0, 1.0/np.sqrt(3)])
        q2 = 2.0*np.pi*np.asarray([0.5, 1.0/np.sqrt(3)/2])
        q3 = 2.0*np.pi*np.asarray([0.5, -1.0/np.sqrt(3)/2])
        mx, my, mz = [], [], []
        for n in range(len(X)):
            R = np.asarray([X[n], Y[n]])
            mx.append(A*np.cos(np.dot(q1,R)))
            my.append(B*np.cos(np.dot(q2,R)))
            mz.append(C*np.cos(np.dot(q3,R)))
        return mx,my,mz


#==============================================================
# takes parameters and arrays X,Y and returns diskret magnetisation
# for skyrmions with corresponding structure
#===============================================================
class Magnetisation_Domainwall():
    def __init__(self, X, Y, r0, direction, width, heli):
        self.mx, self.my, self.mz = self.build_domainwall(X, Y, r0, direction, width, heli)

    def build_domainwall(self, X, Y, r0, direction, width, heli):
        mx, my, mz = [], [], []
        direction = np.asarray(direction)/np.linalg.norm(direction)
        R0 = np.absolute(direction)*r0
        # construct axis for rotation:
        axismat = rotation_matrix([0.0, 0.0, 1.0], heli)
        axis = rotate_spins([direction[0]], [direction[1]], [0.0], axismat)
        axis = np.asarray([axis[0][0], axis[1][0], axis[2][0]])
        # rotate every spin with own matrix according to @theta and @axis
        # the angle theta is given by the DW profile
        for n in range(len(X)):
            r = np.asarray([X[n], Y[n]])
            arg = np.dot(r-R0, direction)/width
            theta = 2.0*np.arctan(np.exp(arg))
            mat = rotation_matrix(axis, theta)
            mag = rotate_spins([0.0] ,[0.0], [1.0], mat)
            mx.append(mag[0][0])
            my.append(mag[1][0])
            mz.append(mag[2][0])
        return mx, my, mz




#==========================================
# LEVEL +1 ROUTINES
# (TAKE @SpinLattice and also return @SpinLattice)
#==========================================

#==============================================
# method for deviding AFM-Lattice in two sublattices
#==============================================
def afm_sublattices(sl):
    X1,Y1,Z1,mx1,my1,mz1,magmom1 = [],[],[],[],[],[],[]
    X2,Y2,Z2,mx2,my2,mz2,magmom2 = [],[],[],[],[],[],[]
    #seperate every second spin and write it in to a sublattice
    for n in range(len(sl.X)):
        if (n%2 + n//sl.size)%2 == 0:
            X1.append(sl.X[n])
            Y1.append(sl.Y[n])
            Z1.append(sl.Z[n])
            mx1.append(sl.mx[n])
            my1.append(sl.my[n])
            mz1.append(sl.mz[n])
            magmom1.append(sl.magmom[n])
        else:
            X2.append(sl.X[n])
            Y2.append(sl.Y[n])
            Z2.append(sl.Z[n])
            mx2.append(sl.mx[n])
            my2.append(sl.my[n])
            mz2.append(sl.mz[n])
            magmom2.append(sl.magmom[n])
    # make them seperate @SpinLattice objects
    sl1 = SpinLattice(size=int(np.sqrt(len(X1))), key='quad', magmom=1.)
    sl2 = SpinLattice(size=int(np.sqrt(len(X2))), key='quad', magmom=1.)
    sl1.X, sl1.Y, sl1.Z = X1, Y1, Z1
    sl1.mx, sl1.my, sl1.mz, sl1.magmom = mx1, my1, mz1, magmom1
    sl2.X, sl2.Y, sl2.Z = X2, Y2, Z2
    sl2.mx, sl2.my, sl2.mz, sl2.magmom = mx2, my2, mz2, magmom2

    return sl1, sl2



# -*- coding: utf-8 -*-
r"""
This module contains the class that takes care of inverting an eigenvector. This is useful in modefollowing approaches
where one wants to investigate both branches of the parabola.
"""
from python3.ioutils.loading_utilities import load_vector_by_name
from python3.ioutils.writing_utilities import write_vector
from typing import TypeVar, List
from pathlib import Path

PathLike = TypeVar("PathLike", str, Path)
import numpy as np


class CInvertEigenvector:
    r"""
    Class for inverting an eigenvector. By means of the htst one can multiply an eigenvector with -1 and get the
    inverted one
    """

    def __init__(self, vector_original: PathLike) -> None:
        r"""
        Initializes the inverting of an eigenvector

        Args:
            vector_original (PathLike): Original eigenvector which shall be inverted.
        """
        self._ev_original = load_vector_by_name(vector_original)

    def __call__(self, outfile: PathLike) -> None:
        r"""
        calls the inverting. Writes out to file:

        Args:
            outfile(PathLike): file to which the inverted eigenvector is written
        """
        write_vector(self._invert_vector(self._ev_original), outfile=outfile)

    @classmethod
    def _invert_vector(cls, vec: List[np.ndarray]) -> List[np.ndarray]:
        r"""
        inverts eigenvector

        Args:
            vec(List[np.ndarray]): eigenvector in format [x, y, z, vx, vy, vz, |v|]

        Returns:
            inverted eigenvector: [x, y, z, -vx, -vy, -vz, |v|]
        """
        return [vec[0], vec[1], vec[2], -1 * vec[3], -1 * vec[4], -1 * vec[5], vec[6]]



# -*- coding: utf-8 -*-
r"""
Module contains functions:
- load eigenvector by number and state
- load specific eigenvector by its filename
"""
from pathlib import Path
from typing import List, TypeVar
import numpy as np

PathLike = TypeVar("PathLike", str, Path)


def load_vector_by_number(path: PathLike, state: str = 'sp', number: int = 1) -> List[np.ndarray]:
    r"""
    loads a specific eigenvector

    Args:
        path(Path): path of directory for the eigenvector
        state(str): String decider whether initial state (ini), saddlepoint state (sp) or final state (fin) should
        be addressed.
        number(int): number of the eigenvector. Default is the first eigenvector

    Returns:
        A list of numpy arrays: x,y,z,vecx,vecy,vecz representing the eigenvector
    """
    if number >= 10:
        filename = 'vector_' + state + '.dat_00' + str(number) + '.dat'
    else:
        filename = 'vector_' + state + '.dat_000' + str(number) + '.dat'
    file = open(path / filename, 'r')
    x, y, z = [], [], []
    vec_x, vec_y, vec_z, v = [], [], [], []
    for line in file:
        L = line.split()
        x.append(float(L[0]))
        y.append(float(L[1]))
        z.append(float(L[2]))
        vec_x.append(float(L[3]))
        vec_y.append(float(L[4]))
        vec_z.append(float(L[5]))
        v.append(float(L[6]))
    return [np.array(x), np.array(y), np.array(z), np.array(vec_x), np.array(vec_y), np.array(vec_z), np.array(v)]


def load_vector_by_name(path: PathLike) -> List[np.ndarray]:
    r"""
    loads a specific eigenvector

    Args:
        path(Path): path ot the specific eigenvector

    Returns:
        A list of numpy arrays: x,y,z,vecx,vecy,vecz, v representing the eigenvector
    """
    file = open(path, 'r')
    x, y, z = [], [], []
    vec_x, vec_y, vec_z, v = [], [], [], []
    for line in file:
        L = line.split()
        x.append(float(L[0]))
        y.append(float(L[1]))
        z.append(float(L[2]))
        vec_x.append(float(L[3]))
        vec_y.append(float(L[4]))
        vec_z.append(float(L[5]))
        v.append(float(L[6]))
    return [np.array(x), np.array(y), np.array(z), np.array(vec_x), np.array(vec_y), np.array(vec_z), np.array(v)]



# -*- coding: utf-8 -*-
r"""
Module for classes analyzing the differences between properties of the magnetizations of the different layers of a
multilayer system.
"""
from pathlib import Path
from typing import TypeVar, Union
import pandas as pd
import numpy as np
import os

PathLike = TypeVar("PathLike", str, Path)


class CMagDiffCubicLayers:
    r"""
    this class is responsible for calculating the differences between two cubic stacked layers. This can be used to
    measure how simultaneous a collapse might be.
    """

    def __init__(self, allspinfiles: bool = True, spinfileslocation: Path = Path.cwd(),
                 specificfile: PathLike = 'spin_0001.dat') -> None:
        r"""
        initializes the calculation of the magnetization deviations in both layers.

        Args:
            allspinfiles(bool): Flag whether all spin_xxx.dat files shell be read and analyzed. The default is True. In this
            case the input parameter vtfileslocation determines where the files should be.

            spinfileslocation(PathLike): All the spin files. Default path is current working directory

            specificfile(PathLike): Only necessary if allspinfiles is False. Then a specific spin-file is analyzed. The
            default is spin_0001.dat here.
        """
        self._allspinfiles = allspinfiles
        self._spinfileslocation = spinfileslocation
        self._specificfile = specificfile

    def __call__(self) -> Union[float, np.ndarray]:
        r"""
        Calls the calculation

        Returns:
            either a float representing the difference of a configuration of a specific file or a 2d numpy array with
            the following structure [[step1, diff1],[step2,diff2],...]
        """
        if self._allspinfiles:
            l_diff, stepindices = [], []
            l_files = self._spinfileslocation.glob('*.dat')
            for f in l_files:
                if str(f.name).startswith('spin_'):
                    stepindices.append(int(str(f.name)[5:-4]))
                    df = pd.read_csv(f, sep=r"\s+", index_col=False, usecols=[0, 1, 2, 3, 4, 5],
                                     names=['x', 'y', 'z', 'sx', 'sy', 'sz'])
                    l_df_lower = df['z'] == 0.0
                    l_df_upper = df['z'] != 0.0
                    diffx = l_df_lower['sx'].to_numpy() - l_df_upper['sx'].to_numpy()
                    diffy = l_df_lower['sy'].to_numpy() - l_df_upper['sy'].to_numpy()
                    diffz = l_df_lower['sz'].to_numpy() - l_df_upper['sz'].to_numpy()
                    vec_diff = np.column_stack((diffx, diffy, diffz))
                    l_diff.append(np.sum(np.linalg.norm(vec_diff, axis=1)))
            l_diff = np.asarray(l_diff)
            l_steps = np.asarray(stepindices)
            return np.column_stack((l_steps, l_diff))
        else:
            df = pd.read_csv(self._specificfile, sep=r"\s+", index_col=False, usecols=[0, 1, 2, 3, 4, 5],
                             names=['x', 'y', 'z', 'sx', 'sy', 'sz'])
            l_df_lower = df['z'] == 0.0
            l_df_upper = df['z'] != 0.0
            diffx = l_df_lower['sx'].to_numpy() - l_df_upper['sx'].to_numpy()
            diffy = l_df_lower['sy'].to_numpy() - l_df_upper['sy'].to_numpy()
            diffz = l_df_lower['sz'].to_numpy() - l_df_upper['sz'].to_numpy()
            vec_diff = np.column_stack((diffx, diffy, diffz))
            return np.sum(np.linalg.norm(vec_diff, axis=1))



# -*- coding: utf-8 -*-
r"""
Lib for python - shell communication

Moritz Goerzen
08.2018
Spintronic Theory Kiel
---------------------------------
OS-System Dependency added by Hendrik Schrautzer
"""
import subprocess
import os
import sys
import platform
from contextlib import contextmanager
import time
from tempfile import mkstemp
from shutil import move, copy2
from os import fdopen, remove
from datetime import datetime

import python3.constants as const
from typing import TypeVar
from pathlib import Path

PathLike = TypeVar("PathLike", str, Path)

# operating system
OPERATINGSYSTEM = platform.system()


@contextmanager
def change_directory(newdir: PathLike) -> None:
    r"""
    Changes the directory using context manager. Will automatically put everything before the yield statement into the
    __enter__-method and everythin behind in the __exit__method.

    Args:
        newdir(str): name of the directory. Has to be a subdirectory of the actual directory.
    """
    prevdir = Path.cwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)


def make_directory(directory: PathLike) -> None:
    r"""
    Creates directory. Chooses the command based on OS. It is possible to create multiple nested directories. It is
    highly recommended to use the input type of pathlibs Path then. Otherwise one has to set the correct back- or
    forward slashes. This destroys the benefit of this platform indepent implementation.

    Args:
        directory(PathLike): new directory
    """
    if OPERATINGSYSTEM == 'Windows':
        command = "md " + str(directory)
        process = subprocess.Popen(command.split(), stdout=subprocess.PIPE, shell=True)
    elif OPERATINGSYSTEM == 'Linux':
        command = "mkdir " + str(directory)
        process = subprocess.Popen(command.split(), stdout=subprocess.PIPE)
    else:
        raise Exception('operating system not yet coded.')
    output, error = process.communicate()


def copy_folder(target: PathLike, destination: PathLike) -> None:
    r"""
    Copies folder from target to destination. Chooses the command based on Os. For windows the /i flag supresses asking
    whether copying and the /e flag enables empty copying.
    """
    if OPERATINGSYSTEM == 'Windows':
        command = "xcopy " + str(target) + " " + str(destination) + "/i /e"
        process = subprocess.Popen(command.split(), stdout=subprocess.PIPE, shell=True)
    elif OPERATINGSYSTEM == 'Linux':
        command = "cp -r " + str(target) + " " + str(destination)
        process = subprocess.Popen(command.split(), stdout=subprocess.PIPE)
    else:
        raise Exception('operating system not yet coded.')
    output, error = process.communicate()


def copy_element(target: PathLike, destination: PathLike) -> None:
    r"""
    Copies an element (target) to a destination. Chooses the command based on OS. On windows the shell=True flag
    must be set. Should be replaced someday by shutil.which
    (https://stackoverflow.com/questions/3022013/windows-cant-find-the-file-on-subprocess-call)

    Args:
        target(PathLike): target file
        destination(PathLike): destination file
    """
    if OPERATINGSYSTEM == 'Windows':
        command = "copy " + str(target) + " " + str(destination)
        process = subprocess.Popen(command.split(), stdout=subprocess.PIPE, shell=True)
    elif OPERATINGSYSTEM == 'Linux':
        command = "cp " + str(target) + " " + str(destination)
        process = subprocess.Popen(command.split(), stdout=subprocess.PIPE)
    else:
        raise Exception('operating system not yet coded.')
    output, error = process.communicate()


def copy_all_elements(targetdirectory, extension, destination):
    # ==================================================================
    for filename in os.listdir(targetdirectory):
        if filename.endswith(extension):
            pathname = os.path.join(targetdirectory, filename)
            if os.path.isfile(pathname):
                copy2(pathname, destination)


def remove_element(filename: PathLike) -> None:
    r"""
    Removes element (filename). Chooses the command based on Os. The windows /q flag is for silent removing. If the file
    is given as a path it is highly recomm. to use pathlibs Path for platform indepent input.
    """
    if OPERATINGSYSTEM == 'Windows':
        command = "del " + str(filename) + " /q"
        process = subprocess.Popen(command.split(), stdout=subprocess.PIPE, shell=True)
    elif OPERATINGSYSTEM == 'Linux':
        command = "rm " + str(filename)
        process = subprocess.Popen(command.split(), stdout=subprocess.PIPE)
    else:
        raise Exception('operating system not yet coded.')
    output, error = process.communicate()


def remove_folder(directory: PathLike) -> None:
    r"""
    Removes directory with everything inside. Chooses the command based on Os. The windows /s flag is for
    deleting subfolders and files and the /q flag is for silent modus.
    """
    if OPERATINGSYSTEM == 'Windows':
        command = "rmdir " + str(directory) + " /s /q"
        process = subprocess.Popen(command.split(), stdout=subprocess.PIPE, shell=True)
    elif OPERATINGSYSTEM == 'Linux':
        command = "rm -r " + str(directory)
        process = subprocess.Popen(command.split(), stdout=subprocess.PIPE)
    else:
        raise Exception('operating system not yet coded.')
    output, error = process.communicate()


def call_bash(scriptname):
    # ==================================================================
    # calls an .sh-script on linux desktop
    bashCommand = "bash " + str(scriptname)
    process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()


def call_sbatch(scriptname):
    # ==================================================================
    # calls an .sh-script on linux clusters
    bashCommand = "sbatch " + str(scriptname)
    process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()


def call_python(scriptname):
    # ==================================================================
    # calls an external python script
    bashCommand = "python " + str(scriptname)
    process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()


def adjust_parameter(keyword, value, directory_file):
    # ==================================================================
    # looks up "directory_file" for the given "keyword" and substitutes everything behind it
    # with "value"
    fh, abs_path = mkstemp()
    with fdopen(fh, 'w') as new_file:
        with open(directory_file) as old_file:
            for line in old_file:
                words = line.split()
                if not words:
                    pass
                elif words[0] == keyword:
                    new_file.write(line.replace(line, keyword + " " + str(value) + "\n"))
                else:
                    new_file.write(line)
    # Remove original file
    remove(directory_file)
    # Move new file
    move(abs_path, directory_file)


def set_jobname(name, directory_file):
    fh, abs_path = mkstemp()
    with fdopen(fh, 'w') as new_file:
        with open(directory_file) as file:
            for line in file:
                if line.startswith('#SBATCH'):
                    if line.split()[1].startswith('--job-name='):
                        new_file.write(line.replace(line, '#SBATCH --job-name=' + name + "\n"))
                    else:
                        new_file.write(line)
                else:
                    new_file.write(line)
    # Remove original file
    remove(directory_file)
    # Move new file
    move(abs_path, directory_file)


def adjust_line_under_key(keyword, value, directory_file):
    # ==================================================================
    # looks up "directory_file" for the given "keyword" and substitutes everything behind it
    # with "value"
    fh, abs_path = mkstemp()
    with fdopen(fh, 'w') as new_file:
        with open(directory_file) as old_file:
            for line in old_file:
                words = line.split()
                if not words:
                    pass
                elif words[0] == keyword:
                    new_file.write(line)
                    line2 = next(old_file)
                    new_file.write(line2.replace(line2, str(value) + "\n"))
                else:
                    new_file.write(line)
    # Remove original file
    remove(directory_file)
    # Move new file
    move(abs_path, directory_file)


def adjust_parameter_new_version(keyword, value, directory_file):
    # ==================================================================
    # looks up "directory_file" for the given "keyword" and substitutes everything behind it
    # with "value"
    keyword_array = keyword.split()
    fh, abs_path = mkstemp()
    with fdopen(fh, 'w') as new_file:
        with open(directory_file) as old_file:
            for line in old_file:
                words = line.split()
                # keywords can consist of more than one word. Every word has to match the corresponding element of the current line
                keyword_match = True
                set_space = True
                if len(keyword_array) == 0 or len(keyword_array) > len(words):
                    keyword_match = False
                else:
                    for i in range(0, len(keyword_array)):
                        if keyword_array[i] not in words[i]:
                            keyword_match = False
                        elif keyword_array[i] != words[i]:
                            set_space = False
                if not words:
                    pass
                elif keyword_match == True:
                    if set_space == True:
                        new_file.write(line.replace(line, keyword + " " + str(value) + "\n"))
                    else:
                        new_file.write(line.replace(line, keyword + str(value) + "\n"))
                else:
                    new_file.write(line)
    # Remove original file
    remove(directory_file)
    # Move new file
    move(abs_path, directory_file)


def make_file(filename):
    # ==================================================================
    # simple touch
    bashCommand = "touch " + str(filename)
    process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()


def call_spin(pathtoexecutable: PathLike = None) -> None:
    r"""
    Calls the spin D algorithm

    Args:
        pathtoexecutable(PathLike, Optional): the path to the SpinD executable. The default is None. In that case the
        default Linux or Windows paths is chosen based on the OS.
    """
    if pathtoexecutable is None:
        if OPERATINGSYSTEM == 'Windows':
            command = const.SPIND_WINDOWS
        elif OPERATINGSYSTEM == 'Linux':
            command = const.SPIND_LINUX
        else:
            raise Exception('operating system not yet coded.')
    else:
        command = pathtoexecutable
    with open('job.out', 'a') as f:
        process = subprocess.Popen(command.split(), stdout=f, shell=True)
    output, error = process.communicate()


def backuping_mkdir(directory):
    # ==================================================================
    # copies folder to oldVersion if already exists and then creates new one as required
    if os.path.isdir(directory):
        temp_directory = directory + "_backup"
        i = 1
        while os.path.isdir(temp_directory):
            temp_directory = directory + "_backup" + str(i)
            i += 1
        print(
            'directory ' + '\"' + directory + '\" ' + 'already exists : backuping old version to ' + '\"' + temp_directory + '\"')
        copy_folder(directory, temp_directory)
        remove_folder(directory)
    make_directory(directory)
    time.sleep(1.0)


def backuping_touch(filename):
    # ==================================================================
    # copies file to oldVersion if already exists and then creates new one as required
    if Path(filename).is_file():
        temp_file = filename + "_backup"
        i = 1
        while Path(temp_file).is_file():
            temp_file = filename + "_backup" + str(i)
            i += 1
        print(
            'file ' + '\"' + filename + '\" ' + 'already exists : backuping old version to ' + '\"' + temp_file + '\"')
        copy_element(filename, temp_file)
        remove_element(filename)
    make_file(filename)
    time.sleep(1.0)


def cluster():
    # ==================================================================
    # reads the current machines name and returns the bool variable @cluster
    bashCommand = 'hostname'
    process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
    output, error = process.communicate()
    # ADD INDICATOR FOR YOUR MACHINE NAME HERE
    # OUTPUT IS IN FORMAT bytestream SO USE @b IN FRONT OF STRING
    desktop_machine_names = [b'goerzen', b'sh8', b'DESKTOP-TIBLOAM']
    for name in desktop_machine_names:
        if output.startswith(name):
            return False
    # MACHINE IS CONSIDERED TO BE SLURM CLUSTER IF NAMES DID NOT MATCH
    return True


def read_last_line(filename):
    # ==================================================================
    num_lines = sum(1 for line in open(filename))
    with open(filename, 'rb') as f:
        if num_lines > 1:
            f.seek(-2, os.SEEK_END)
            while f.read(1) != b'\n':
                f.seek(-2, os.SEEK_CUR)
            return f.readline().decode()
        else:
            return f.readline().decode()


def writelog(file: PathLike, message: str, end: str = '\n') -> None:
    r"""
    During testing one wants to show the result of the test within the console but also write a log file presenting
    the results of the tests for later usage.

    Args:
        file(PathLike): log file

        message(str): message which is printed as well as written to the logfile

        end(str): indicator for the end of the lie
    """
    print(message, end=end)
    with open(file, 'a') as f:
        f.write(message + end)


def startlogger(file: PathLike) -> None:
    r"""
    Starts the logging.

    Args:
        file(PathLike): log file
    """
    if os.path.isfile(file):
        remove_element(file)
    writelog(file=file, message='#' * 100)
    writelog(file=file, message=f'Test were run at: {datetime.now()}')
    writelog(file=file, message='#' * 100)



# -*- coding: utf-8 -*-
r"""
This script starts the test procedures for multilayer. Make sure that you are using the correct version of the SpinD-
dynamics code since multilayer are relatively new (~2019).
"""
from tests.test_multilayers.test_precission_withoutdamping.ctestprecisionwithoutdamping import \
    CTestPrecisionWithoutDamping
from python3.shell_commands import writelog, startlogger
from pathlib import Path


def main() -> None:
    r"""
    Main test routine:

    - Test precision frequency for two spins from different layers coupled via different types of interlayer exchange:
        [ab, bc, ac, AA, AB, AC]
    """
    interlayerinteraction=['ab', 'bc', 'ac', 'AA', 'AB', 'AC', 'BA', 'BB', 'BC', 'CA', 'CB', 'CC']

    logfile = Path.cwd() / 'multilayer_tests.log'
    startlogger(logfile)
    writelog(logfile, 'Perform multilayer testing: ....')
    writelog(logfile, '#' * 100)
    writelog(logfile,
             'Testing precision of two spins from different layers coupled via interlayer ex. (without damping):')
    for inter in interlayerinteraction:
        writelog(logfile, '=' * 100)
        l_test = CTestPrecisionWithoutDamping(logfile=logfile, cleanup=True, interlayerinteraction=inter)
        writelog(logfile, message=repr(l_test))
        result = l_test()


if __name__ == "__main__":
    main()



import sys
import os
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import MultipleLocator, NullFormatter
from mpl_toolkits.axes_grid1.inset_locator import (inset_axes, InsetPosition, mark_inset)
from python3.magnetisations import SpinLattice
from python3.inputs_simulation import InputPhys
from python3.routines_analytic import *
from python3.visualizations import *
from python3.shell_commands import *
from python3.spinD_kiel_jobs import *
from pathlib import Path


def test_ssdispersion(inputs):
    # global settings
    para = InputPhys(file=inputs / 'inp')
    size = 100
    key = 'hex'
    magmom = 1.0
    if not os.path.isdir('test_calculations'):
        make_directory('test_calculations')
    if not os.path.isdir('test_calculations/GM'):
        make_directory('test_calculations/GM')
    if not os.path.isdir('test_calculations/GK'):
        make_directory('test_calculations/GK')

    num_energies = []
    ana_energies = []
    dif_energies = []
    qq = []
    qs = np.linspace(0.0, 0.49, 50)

    # GM-direction
    Rq = [0.0, 0.0, 1.0]
    Iq = [1.0, 0.5, 0.0]
    for q in qs:
        # begin with lowest q
        q = round(max(qs) - q, 2)
        sl = SpinLattice(size=size, key=key, path=None, magmom=magmom)
        Q = np.asarray([1.0, 0.0, 0.0]) * q
        sl.add1qstate(Rq, Iq, -Q)
        num = numeric_energy(sl, Path.cwd() / 'test_calculations' / 'GM'/ ('image_q_' + str(q)), inputs)
        ana = energy_SS(-q * 2. / 3. ** .5, para)
        num_energies.append(num)
        ana_energies.append(ana)
        # dif_energies.append(ana-num)
        qq.append(-q * 2. / 3. ** .5)

    # GK-direction
    Rq = [0.0, 0.0, 1.0]
    Iq = [1.0, 1.0, 0.0]
    for q in qs:
        q = round(q, 2)
        sl = SpinLattice(size=size, key=key, path=None, magmom=magmom)
        Q = np.asarray([1.0, 1.0, 0.0]) * q
        sl.add1qstate(Rq, Iq, -Q)
        num = numeric_energy(sl, 'test_calculations/GK/image_q_' + str(q), inputs)
        ana = energy_SS(q * 2., para)
        num_energies.append(num)
        ana_energies.append(ana)
        # dif_energies.append(ana-num)
        qq.append(q * 2.)

    # if any(dif_energies[0][n] <= 1e-9 for n in range(len(qs))) :
    #    print '====================================================='
    #    print ' ERROR: discrepancy of numerical and analytic results'
    #    print '        for the spinspiral dispersion exceeded < 1e-9'
    #    print '====================================================='

    return qq, np.asarray(ana_energies).transpose(), np.asarray(
        num_energies).transpose()  # , np.asarray(dif_energies).transpose()


def plot_ssdispersion(inputs, qs, ana, num):
    para = InputPhys(file=inputs + '/inp')
    FM = energy_SS(0, para)

    # build figure
    fig = plt.figure()
    fig.set_size_inches(8, 8)
    ax0 = plt.subplot2grid((5, 2), (0, 0), rowspan=3)
    ax1 = plt.subplot2grid((5, 2), (3, 0))
    ax2 = plt.subplot2grid((5, 2), (4, 0))
    ax0e = plt.subplot2grid((5, 2), (0, 1), rowspan=3)
    ax1e = plt.subplot2grid((5, 2), (3, 1))
    ax2e = plt.subplot2grid((5, 2), (4, 1))
    plt.subplots_adjust(wspace=0.001, hspace=0.05, left=0.11, right=0.84)

    # add labels
    ax2.set_xlabel(r'q / $\frac{2\pi}{a}$', fontsize=20)
    ax0.set_ylabel(r'$E(q)-E_{FM}$ /(meV/atom)', fontsize=20)

    # make brillouin-like xaxis
    ax0.yaxis.set_minor_locator(MultipleLocator(10))
    ax0.xaxis.set_minor_locator(MultipleLocator(0.1))
    ax0.xaxis.set_major_formatter(NullFormatter())
    ax0.xaxis.set_minor_formatter(NullFormatter())
    ax0e.xaxis.set_major_formatter(NullFormatter())
    ax0e.xaxis.set_minor_formatter(NullFormatter())
    ax0.xaxis.set_tick_params(labelsize=15, direction='in', which='both')
    ax0.yaxis.set_tick_params(labelsize=15, direction='in', which='both')
    ax0e.xaxis.set_tick_params(labelsize=15, direction='in', which='both')
    ax0e.yaxis.set_tick_params(labelsize=15, direction='in', which='both')
    ax1.yaxis.set_minor_locator(MultipleLocator(10))
    ax1.xaxis.set_minor_locator(MultipleLocator(0.1))
    ax1e.xaxis.set_minor_locator(MultipleLocator(0.1))
    ax1.xaxis.set_major_formatter(NullFormatter())
    ax1.xaxis.set_minor_formatter(NullFormatter())
    ax1e.xaxis.set_major_formatter(NullFormatter())
    ax1e.xaxis.set_minor_formatter(NullFormatter())
    ax2.yaxis.set_minor_locator(MultipleLocator(10))
    ax2.xaxis.set_minor_locator(MultipleLocator(0.1))
    ax1.xaxis.set_tick_params(labelsize=15, direction='in', which='both')
    ax1.yaxis.set_tick_params(labelsize=15, direction='in', which='both')
    ax2.xaxis.set_tick_params(labelsize=15, direction='in', which='both')
    ax2.yaxis.set_tick_params(labelsize=15, direction='in', which='both')
    ax1e.xaxis.set_tick_params(labelsize=15, direction='in', which='both')
    ax1e.yaxis.set_tick_params(labelsize=15, direction='in', which='both')
    ax2e.xaxis.set_tick_params(labelsize=15, direction='in', which='both')
    ax2e.yaxis.set_tick_params(labelsize=15, direction='in', which='both')
    ax2.set_xticks(np.array([-0.577, 0, 0.666, 1]))
    ax2.set_xticklabels(['M', '$\Gamma$', 'K', 'M'])
    ax2e.set_xticks(np.array([-0.577, 0, 0.666, 1]))
    ax2e.set_xticklabels(['M', '$\Gamma$', 'K', 'M'])

    ax0.set_xlim(-0.577, 1.0)
    ax1.set_xlim(-0.577, 1.0)
    ax2.set_xlim(-0.577, 1.0)

    ax0.plot(qs, (ana[0] - FM[0]) * 1000.0, c='b', ls='-', marker='')
    ax0.plot(qs, (num[0] - num[0][50]) * 1000.0, c='b', ls='', marker='+')
    ax1.plot(qs, (ana[4] - FM[4]) * 1000.0, c='r', ls='-', marker='')
    ax1.plot(qs, (num[4] - num[4][50]) * 1000.0, c='r', ls='', marker='+')
    ax1.plot(qs, (ana[3] - FM[3]) * 1000.0, c='m', ls='-', marker='')
    ax1.plot(qs, (num[3] - num[3][50]) * 1000.0, c='m', ls='', marker='+')
    ax2.plot(qs, (ana[5] - FM[5]) * 1000.0, c='lime', ls='-', marker='')
    ax2.plot(qs, (num[5] - num[5][50]) * 1000.0, c='lime', ls='', marker='+')
    ax2.plot(qs, (ana[6] - FM[6]) * 1000.0, c='g', ls='-', marker='')
    ax2.plot(qs, (num[6] - num[6][50]) * 1000.0, c='g', ls='', marker='+')
    ax2.plot(qs, (ana[7] - FM[7]) * 1000.0, c='k', ls='-', marker='')
    ax2.plot(qs, (num[7] - num[7][50]) * 1000.0, c='k', ls='', marker='+')

    ax0e.plot(qs, ana[0] - FM[0] - num[0] + num[0][50], c='b', marker='+', ls='', label=r'E$_{tot}$')
    ax0e.plot(qs, ana[1] - FM[1] - num[1] + num[1][50], c='g', marker='+', ls='', label=r'E$_{exc}$')
    ax1e.plot(qs, ana[4] - FM[4] - num[4] + num[4][50], c='r', marker='+', ls='', label=r'E$_{dmi}$')
    ax1e.plot(qs, ana[3] - FM[3] - num[3] + num[3][50], c='m', marker='+', ls='', label=r'E$_{ani}$')
    ax2e.plot(qs, ana[5] - FM[5] - num[5] + num[5][50], c='lime', marker='+', ls='', label=r'E$_{3sp}$')
    ax2e.plot(qs, ana[6] - FM[6] - num[6] + num[6][50], c='g', marker='+', ls='', label=r'E$_{4sp}$')
    ax2e.plot(qs, ana[7] - FM[7] - num[7] + num[7][50], c='k', marker='+', ls='', label=r'E$_{biq}$')

    # draw borders
    ax0.arrow(0.0, ax0.get_ylim()[0], 0.0, ax0.get_ylim()[1], head_width=0, head_length=0, color='k', ls='-',
              width=0.00002)
    ax0e.arrow(0.0, ax0e.get_ylim()[0], 0.0, ax0e.get_ylim()[1], head_width=0, head_length=0, color='k', ls='-',
               width=0.00002)
    ax0.arrow(0.666, ax0.get_ylim()[0], 0.0, ax0.get_ylim()[1], head_width=0, head_length=0, color='k', ls='-',
              width=0.00002)
    ax0e.arrow(0.666, ax0e.get_ylim()[0], 0.0, ax0e.get_ylim()[1], head_width=0, head_length=0, color='k', ls='-',
               width=0.00002)
    ax0.arrow(-0.577, 0.0, 0.577 + 1, 0.0, head_width=0, head_length=0, color='k', ls='-', width=0.00002)
    ax1.arrow(0.0, ax1.get_ylim()[0], 0.0, ax1.get_ylim()[1], head_width=0, head_length=0, color='k', ls='-',
              width=0.00002)
    ax1e.arrow(0.0, ax1e.get_ylim()[0], 0.0, ax1e.get_ylim()[1], head_width=0, head_length=0, color='k', ls='-',
               width=0.00002)
    ax1.arrow(0.666, ax1.get_ylim()[0], 0.0, ax1.get_ylim()[1], head_width=0, head_length=0, color='k', ls='-',
              width=0.00002)
    ax1e.arrow(0.666, ax1e.get_ylim()[0], 0.0, ax1e.get_ylim()[1], head_width=0, head_length=0, color='k', ls='-',
               width=0.00002)
    ax1.arrow(-0.577, 0.0, 0.577 + 1, 0.0, head_width=0, head_length=0, color='k', ls='-', width=0.00002)
    ax2.arrow(0.0, ax2.get_ylim()[0], 0.0, ax2.get_ylim()[1], head_width=0, head_length=0, color='k', ls='-',
              width=0.00002)
    ax2.arrow(0.666, ax2.get_ylim()[0], 0.0, ax2.get_ylim()[1], head_width=0, head_length=0, color='k', ls='-',
              width=0.00002)
    ax2e.arrow(0.0, ax2e.get_ylim()[0], 0.0, ax2e.get_ylim()[1], head_width=0, head_length=0, color='k', ls='-',
               width=0.00002)
    ax2e.arrow(0.666, ax2e.get_ylim()[0], 0.0, ax2e.get_ylim()[1], head_width=0, head_length=0, color='k', ls='-',
               width=0.00002)
    ax2.arrow(-0.577, 0.0, 0.577 + 1, 0.0, head_width=0, head_length=0, color='k', ls='-', width=0.00002)

    ax0e.yaxis.tick_right()
    ax1e.yaxis.tick_right()
    ax2e.yaxis.tick_right()

    # color the area beneath 0
    ulim0 = ax0.get_ylim()[0]
    ax0.fill_between(qs, [ulim0 for i in range(len(qs))], color='k', alpha=0.1)
    ulim1 = ax1.get_ylim()[0]
    ax1.fill_between(qs, [ulim1 for i in range(len(qs))], color='k', alpha=0.1)
    ulim2 = ax2.get_ylim()[0]
    ax2.fill_between(qs, [ulim2 for i in range(len(qs))], color='k', alpha=0.1)

    ax0e.legend(bbox_to_anchor=(0, 1), loc=2, borderaxespad=0., fontsize=8)
    ax1e.legend(bbox_to_anchor=(0, 1), loc=2, borderaxespad=0., fontsize=8)
    ax2e.legend(bbox_to_anchor=(0, 1), loc=2, borderaxespad=0., fontsize=8)
    plt.savefig('test_spinspiral_energydispersion.png')
    plt.show()


if __name__ == '__main__':
    inputs = Path.cwd().parent / 'input_data'
    qs, ana, num = test_ssdispersion(inputs)
    plot_ssdispersion(inputs, qs, ana, num)
